import asyncio
import random
from collections import deque
from typing import Any, Iterator

import item as item
from naff import ActiveVoiceState
from naff.api.voice.audio import BaseAudio

__all__ = ("NaffQueue",)


class NaffQueue:
    voice_state: ActiveVoiceState
    """The voice state this queue is playing for."""
    _entries: deque
    """The queue's base data store"""
    _item_queued: asyncio.Event
    """An event to be fired whenever an item is queued."""

    def __init__(self, voice_state: ActiveVoiceState):
        self.voice_state = voice_state

        self._entries = deque()
        self._item_queued = asyncio.Event()

    def __len__(self) -> int:
        return len(self._entries)

    def __iter__(self) -> Iterator[BaseAudio]:
        return iter(self._entries)

    def put(self, audio: BaseAudio) -> None:
        """
        Enqueue audio at the end of the queue.

        Args:
            audio: The audio to enqueue
        """
        self._entries.append(audio)
        self._item_queued.set()

    def put_first(self, audio: BaseAudio) -> None:
        """
        Enqueue Audio to be played next.

        Args:
            audio: The audio to enqueue
        """
        self._entries.appendleft(audio)
        self._item_queued.set()

    async def pop(self) -> BaseAudio:
        """
        Pop the next item from the queue.

        Will wait if there are no items in the queue.
        """
        if len(self) == 0:
            # wait for an item to be enqueued
            await self._item_queued.wait()
        item = self._entries.popleft()
        self._item_queued.clear()
        return item

    def pop_no_wait(self) -> BaseAudio:
        """Pop from the queue without waiting."""
        return self._entries.popleft()

    def shuffle(self) -> None:
        """Shuffle the queue."""
        random.shuffle(self._entries)

    def clear(self) -> None:
        """Clear the queue."""
        self._entries.clear()

    def peek(self, positions: int = 1) -> BaseAudio | None:
        """
        Peek ahead `position` in the queue.

        Args:
            positions: How many positions ahead to peek at.

        Returns:
            BaseAudio if anything at the given position
        """
        try:
            return self._entries[positions - 1]
        except IndexError:
            return None

    def peek_at_index(self, index: int) -> BaseAudio:
        """
        Peek at a specific Index

        Args:
            index: The index to peek at

        Returns:
            BaseAudio at given index
        """
        return self._entries[index]

    async def __playback_queue(self) -> None:
        """The queue task itself. While the vc is connected, it will play through the enqueued audio"""
        while self.voice_state.connected:
            if self.voice_state.playing:
                await self.voice_state.wait_for_stopped()
            audio = await self.pop()
            await self.voice_state.play(audio)

    async def __call__(self) -> None:
        await self.__playback_queue()

    def start(self) -> None:
        """Create a background Queue task"""
        asyncio.create_task(self())



import asyncio

from naff.api.voice.audio import AudioVolume
from typing_extensions import Self
from yt_dlp import YoutubeDL

__all__ = ("YTAudio",)

youtube_dl = YoutubeDL(
    {
        "format": "bestaudio/best",
        "outtmpl": "%(extractor)s-%(id)s-%(title)s.%(ext)s",
        "restrictfilenames": True,
        "noplaylist": True,
        "nocheckcertificate": True,
        "ignoreerrors": False,
        "logtostderr": False,
        "quiet": True,
        "no_warnings": True,
        "default_search": "auto",
        "source_address": "0.0.0.0",  # noqa: S104
    }
)


class YTAudio(AudioVolume):
    def __init__(self, src: str) -> None:
        super().__init__(src)
        self.entry: dict | None = None
        """The audio entry this audio object represents."""

    @classmethod
    async def from_url(
        cls, url: str, stream: bool = True, ytdl: YoutubeDL | None = None
    ) -> Self:
        """
        Create this object from a YTDL supported url

        Args:
            url: The url to play from
            stream: Should this audio be streamed, or downloaded
            ytdl: An optional YoutubeDL like object to use to fetch

        Returns:
            A YTAudio object
        """
        if not ytdl:
            ytdl = youtube_dl

        data = await asyncio.to_thread(
            lambda: ytdl.extract_info(url, download=not stream)
        )

        if "entries" in data:
            data = data["entries"][0]

        filename = data["url"] if stream else ytdl.prepare_filename(data)

        new_cls = cls(filename)

        if stream:
            new_cls.ffmpeg_before_args = (
                "-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5"
            )

        new_cls.entry = data
        return new_cls



from pathlib import Path

from setuptools import find_packages, setup

from naff_audio import __version__

setup(
    name="naff-audio",
    description="A collection of useful audio scripts for NAFF",
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    author="LordOfPolls",
    author_email="lordofpolls.dev@gmail.com",
    url="",
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.10",
    install_requires=(Path(__file__).parent / "requirements.txt")
    .read_text()
    .splitlines(),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Documentation",
        "Typing :: Typed",
    ],
)
